from pydantic import BaseModel
from typing import Optional, List
import typing as T

class BotGuild(BaseModel):
    nickname: str
    tiny_id: int
    avatar_url: str

class Guild(BaseModel):
    guild_id: int
    guild_name: str
    guild_display_id: Optional[int]
    # 以下是 getGuildMetaByGuest 返回的更多结果
    guild_profile: Optional[str]
    create_time: Optional[int]
    max_member_count: Optional[int]
    max_robot_count: Optional[int]
    max_admin_count: Optional[int]
    member_count: Optional[int]
    owner_id: Optional[int]

class SlowMode(BaseModel):
    slow_mode_key: int
    slow_mode_text: str
    speak_frequency: int
    slow_mode_circle: int

class Channel(BaseModel):
    owner_guild_id: int
    channel_id: int
    channel_type: int
    channel_name: str
    create_time: int
    creator_id: int
    creator_tiny_id: int
    talk_permission: int
    visible_type: int
    current_slow_mode: int
    slow_modes: List[SlowMode]

class GuildMember(BaseModel):
    tiny_id: Optional[int]
    user_id: Optional[int]
    title: Optional[str]
    nickname: str
    role: Optional[int]
    # 仅论坛子频道
    icon_url: Optional[str]

class GuildMembers(BaseModel):
    members: List[GuildMember]
    bots: List[GuildMember]
    admins: List[GuildMember]

class Reaction(BaseModel):
    emoji_id: str
    emoji_index: int
    emoji_type: int
    emoji_name: str
    count: int
    clicked: bool

class Role(BaseModel):
    role_id: int
    role_name: Optional[str]
    argb_color: Optional[int]
    independent: Optional[bool]
    member_count: Optional[int]
    max_count: Optional[int]
    owned: Optional[int]
    disabled: Optional[bool]

"""
论坛子频道相关
"""

class TopicChannelFile(BaseModel):
    file_id: str
    pattern_id: str
    url: str
    width: int
    height: int

class TopicChannelFeedResource(BaseModel):
    images: List[TopicChannelFile]
    videos: List[TopicChannelFile]

class TopicChannelFeed(BaseModel):
    id: int
    title: str
    sub_title: str
    create_time: int
    guild_id: int
    channel_id: int
    poster_info: GuildMember
    contents: list # TODO 解析论坛子频道 contents，参考 https://github.com/Mrs4s/go-cqhttp/blob/7278f99ed9118fe2e54aca752e62574d5bae3f00/coolq/feed.go#L10
    resources: TopicChannelFeedResource


import inspect
import os
import typing as T
from collections import namedtuple

from .logger import logger

Parameter = namedtuple("Parameter", ["name", "annotation", "default"])

TRACEBACKED = os.urandom(32)

def raiser(error):
    raise error

def protocol_log(func):
    async def wrapper(*args, **kwargs):
        try:
            result = await func(*args, **kwargs)
            logger.info(f"Protocol: protocol method {func.__name__} was called")
            return result
        except Exception as e:
            logger.error(f"Protocol: protocol method {func.__name__} raised a error: {e.__class__.__name__}")
            raise e

    return wrapper

def argument_signature(callable_target) -> T.List[Parameter]:
    return [
        Parameter(
            name=name,
            annotation=param.annotation if param.annotation != inspect._empty else None,
            default=param.default if param.default != inspect._empty else None
        )
        for name, param in dict(inspect.signature(callable_target).parameters).items()
    ]

import re
from .entities.components import ComponentTypes

class CQParser:
    def __replaceChar(self, string, char, start, end):
        string = list(string)
        del (string[start:end])
        string.insert(start, char)
        return ''.join(string)

    # 获得文本中每一个 CQ 码的起始和结束位置
    def __getCQIndex(self, text):
        cqIndex = []
        for m in re.compile("(\[CQ:(.+?)])").finditer(text):
            cqIndex.append((m.start(), m.end()))
        cqIndex.append((len(text), len(text)))
        return cqIndex

    # 转义中括号
    def escape(self, text, isEscape=True):
        if isEscape:
            text = text.replace("&", "&amp;")
            text = text.replace(",", "&#44;")
            text = text.replace("[", "&#91;")
            text = text.replace("]", "&#93;")
        else:
            text = text.replace("&amp;", "&")
            text = text.replace("&#44;", ",")
            text = text.replace("&#91;", "[")
            text = text.replace("&#93;", "]")
        return text

    # 将纯文本转换成类型为 plain 的 CQ 码
    def plainToCQ(self, text):
        i = j = k = 0
        cqIndex = self.__getCQIndex(text)
        while i < len(cqIndex):
            if i > 0:
                if i == 1:
                    k += 1
                else:
                    j += 1
            cqIndex = self.__getCQIndex(text)
            if i > 0:
                l, r = cqIndex[j][k], cqIndex[j + 1][0]
            else:
                l, r = 0, cqIndex[0][0]
            source_text = text[l:r]
            if source_text != "":
                text = self.__replaceChar(text, f"[CQ:plain,text={self.escape(source_text)}]", l, r)
            i += 1
        return text
    
    def getAttributeList(self, text):
        text_array = text.split(",")
        text_array.pop(0)
        attribute_list = {}
        for _ in text_array:
            regex_result = re.search(r"^(.*?)=([\s\S]+)", _)
            k = regex_result.group(1)
            if k == "type":
                k = "_type"
            v = self.escape(regex_result.group(2), isEscape=False)
            attribute_list[k] = v
        return attribute_list

    def parseChain(self, text):
        text = self.plainToCQ(text)
        cqcode_list = re.findall(r'(\[CQ:([\s\S]+?)])', text)
        chain = []
        for x in cqcode_list:
            message_type = re.search(r"^\[CQ\:(.*?)\,", x[0]).group(1)
            try:
                chain.append(ComponentTypes[message_type].parse_obj(self.getAttributeList(x[1])))
            except:
                chain.append(ComponentTypes["unknown"].parse_obj({"text": message_type}))
                logger.error(f"Protocol: Cannot convert message type: {message_type}")
        return chain


import typing as T
from enum import Enum
from pydantic import BaseModel

from ..entities import Friend, Member, Anonymous, File, OfflineFile, Device, GuildMember, Reaction, Channel
from ..misc import CQParser

parser = CQParser()

class AppInitEvent(BaseModel):
    pass

class MessageItemType(Enum):
    FriendMessage = "FriendMessage"
    GroupMessage = "GroupMessage"
    GuildMessage = "GuildMessage"
    BotMessage = "BotMessage"
    Message = "Message"

class FriendMessage(BaseModel):
    type: MessageItemType = "FriendMessage"
    time: int
    self_id: str
    sub_type: str
    message_id: int
    user_id: int
    message: T.Union[str, list]
    raw_message: str
    font: int
    sender: Friend

    def __init__(self, message: str, **_):
        message = parser.parseChain(message)
        super().__init__(message=message, **_)

class GroupMessage(BaseModel):
    type: MessageItemType = "GroupMessage"
    self_id: int
    sub_type: str
    message_id: int
    group_id: int
    user_id: int
    anonymous: T.Optional[Anonymous]
    message: T.Union[str, list]
    raw_message: str
    font: int
    sender: Member
    time: int

    def __init__(self, message: str, **_):
        message = parser.parseChain(message)
        super().__init__(message=message, **_)

class GuildMessage(BaseModel):
    type: MessageItemType = "GuildMessage"
    self_id: int
    self_tiny_id: int
    sub_type: str
    message_id: str
    guild_id: int
    channel_id: int
    user_id: int
    message: T.Union[str, list]
    sender: GuildMember
    raw_message: T.Optional[str]

    def __init__(self, message: str, **_):
        raw_message = message
        message = parser.parseChain(message)
        super().__init__(message=message, raw_message=raw_message, **_)

class BotMessage(BaseModel):
    type: MessageItemType = "BotMessage"
    message_id: T.Union[int, str]

class Message(BaseModel):  # getMessage
    type: MessageItemType = "Message"
    message_id: int
    real_id: int
    sender: Member
    time: int
    message: str
    raw_message: str

    def __init__(self, message: str, **_):
        message = parser.parseChain(message)
        super().__init__(message=message, **_)

MessageTypes = {
    "private": FriendMessage,
    "group": GroupMessage,
    "guild": GuildMessage
}

class ForwardMessageSender(BaseModel):
    nickname: str
    user_id: int

class ForwardMessageNode(BaseModel):
    content: T.Union[str, list]
    raw_content: T.Optional[str] # 本来没有的，用于表示原 content
    sender: ForwardMessageSender
    time: int

    def __init__(self, content: str, **_):
        raw_content = content
        content = parser.parseChain(content)
        super().__init__(content=content, raw_content=raw_content, **_)

class ForwardMessages(BaseModel):
    messages: T.List[ForwardMessageNode]

class NoticeItemType(Enum):
    GroupFileUpload = "GroupFileUpload"
    GroupAdminChange = "GroupAdminChange"
    GroupMemberDecrease = "GroupMemberDecrease"
    GroupMemberIncrease = "GroupMemberIncrease"
    GroupMemberBan = "GroupMemberBan"
    FriendAdd = "FriendAdd"
    GroupMessageRecall = "GroupMessageRecall"
    FriendMessageRecall = "FriendMessageRecall"
    Notify = "Notify"
    GroupCardChange = "GroupCardChange"
    FriendOfflineFile = "FriendOfflineFile"
    ClientStatusChange = "ClientStatusChange"
    EssenceMessageChange = "EssenceMessageChange"
    # 以下为频道事件
    MessageReactionsUpdated = "MessageReactionsUpdated"
    ChannelUpdated = "ChannelUpdated"
    ChannelCreated = "ChannelCreated"
    ChannelDestroyed = "ChannelDestroyed"
    GuildChannelRecall = "GuildChannelRecall"

class GroupFileUpload(BaseModel):
    type: NoticeItemType = "GroupFileUpload"
    time: int
    self_id: int
    group_id: int
    user_id: int
    file: File

class GroupAdminChange(BaseModel):
    type: NoticeItemType = "GroupAdminChange"
    time: int
    self_id: int
    sub_type: str
    group_id: int
    user_id: int

class GroupMemberDecrease(BaseModel):
    type: NoticeItemType = "GroupMemberDecrease"
    time: int
    self_id: int
    sub_type: str
    group_id: int
    operator_id: int
    user_id: int

class GroupMemberIncrease(BaseModel):
    type: NoticeItemType = "GroupMemberIncrease"
    time: int
    self_id: int
    sub_type: str
    group_id: int
    operator_id: int
    user_id: int

class GroupMemberBan(BaseModel):
    type: NoticeItemType = "GroupMemberBan"
    time: int
    self_id: int
    sub_type: str
    group_id: int
    operator_id: int
    user_id: int
    duration: int

class FriendAdd(BaseModel):
    type: NoticeItemType = "FriendAdd"
    time: int
    self_id: int
    user_id: int

class GroupMessageRecall(BaseModel):
    type: NoticeItemType = "GroupMessageRecall"
    time: int
    self_id: int
    group_id: int
    user_id: int
    operator_id: int
    message_id: int

class FriendMessageRecall(BaseModel):
    type: NoticeItemType = "FriendMessageRecall"
    time: int
    self_id: int
    user_id: int
    message_id: int

class Notify(BaseModel):
    type: NoticeItemType = "Notify"
    sub_type: str
    user_id: int
    target_id: T.Optional[int]
    time: T.Optional[int]
    self_id: T.Optional[int]
    group_id: T.Optional[int]
    honor_type: T.Optional[str]

class GroupCardChange(BaseModel):
    type: NoticeItemType = "GroupCardChange"
    group_id: int
    user_id: int
    card_new: str
    card_old: str

class FriendOfflineFile(BaseModel):
    type: NoticeItemType = "FriendOfflineFile"
    user_id: int
    file: OfflineFile

class ClientStatusChange(BaseModel):
    type: NoticeItemType = "ClientStatusChange"
    client: Device
    online: bool

class EssenceMessageChange(BaseModel):
    type: NoticeItemType = "EssenceMessageChange"
    sub_type: str
    sender_id: int
    operator_id: int
    message_id: int

class MessageReactionsUpdated(BaseModel):
    type: NoticeItemType = "MessageReactionsUpdated"
    guild_id: int
    channel_id: int
    user_id: int
    message_id: str
    current_reactions: T.List[Reaction]

class ChannelUpdated(BaseModel):
    type: NoticeItemType = "ChannelUpdated"
    guild_id: int
    channel_id: int
    user_id: int
    operator_id: int
    old_info: Channel
    new_info: Channel

class ChannelCreated(BaseModel):
    guild_id: int
    channel_id: int
    user_id: int
    operator_id: int
    channel_info: Channel

class ChannelDestroyed(BaseModel):
    guild_id: int
    channel_id: int
    user_id: int
    operator_id: int
    channel_info: Channel

class GuildChannelRecall(BaseModel):
    guild_id: int
    channel_id: int
    operator_id: int
    message_id: str
    user_id: int

NoticeTypes = {
    "group_upload": GroupFileUpload,
    "group_admin": GroupAdminChange,
    "group_decrease": GroupMemberDecrease,
    "group_increase": GroupMemberIncrease,
    "group_ban": GroupMemberBan,
    "friend_add": FriendAdd,
    "group_recall": GroupMessageRecall,
    "friend_recall": FriendMessageRecall,
    "notify": Notify,
    "group_card": GroupCardChange,
    "offline_file": FriendOfflineFile,
    "client_status": ClientStatusChange,
    "essence": EssenceMessageChange,
    "message_reactions_updated": MessageReactionsUpdated,
    "channel_updated": ChannelUpdated,
    "channel_created": ChannelCreated,
    "channel_destroyed": ChannelDestroyed,
    "guild_channel_recall": GuildChannelRecall
}

class RequestItemType(Enum):
    FriendRequest = "FriendRequest"
    GroupRequest = "GroupRequest"

class FriendRequest(BaseModel):
    type: RequestItemType = "FriendRequest"
    time: int
    self_id: int
    user_id: int
    comment: str
    flag: str

class GroupRequest(BaseModel):
    type: RequestItemType = "GroupRequest"
    time: int
    self_id: int
    sub_type: str
    group_id: int
    user_id: int
    comment: str
    flag: str

RequestTypes = {
    "friend": FriendRequest,
    "group": GroupRequest
}


from .models import *
from enum import Enum

class ExternalEvents(Enum):
    AppInitEvent = AppInitEvent

    GroupFileUpload = GroupFileUpload
    GroupAdminChange = GroupAdminChange
    GroupMemberDecrease = GroupMemberDecrease
    GroupMemberIncrease = GroupMemberIncrease
    GroupMemberBan = GroupMemberBan
    FriendAdd = FriendAdd
    GroupMessageRecall = GroupMessageRecall
    FriendMessageRecall = FriendMessageRecall
    Notify = Notify
    GroupCardChange = GroupCardChange
    FriendOfflineFile = FriendOfflineFile
    ClientStatusChange = ClientStatusChange
    EssenceMessageChange = EssenceMessageChange
    
    MessageReactionsUpdated = MessageReactionsUpdated
    ChannelUpdated = ChannelUpdated
    ChannelCreated = ChannelCreated
    ChannelDestroyed = ChannelDestroyed
    GuildChannelRecall = GuildChannelRecall

    FriendRequest = FriendRequest
    GroupRequest = GroupRequest

class ExternalEventTypes(Enum):
    AppInitEvent = "AppInitEvent"

    GroupFileUpload = "GroupFileUpload"
    GroupAdminChange = "GroupAdminChange"
    GroupMemberDecrease = "GroupMemberDecrease"
    GroupMemberIncrease = "GroupMemberIncrease"
    GroupMemberBan = "GroupMemberBan"
    FriendAdd = "FriendAdd"
    GroupMessageRecall = "GroupMessageRecall"
    FriendMessageRecall = "FriendMessageRecall"
    Notify = "Notify"
    GroupCardChange = "GroupCardChange"
    FriendOfflineFile = "FriendOfflineFile"
    ClientStatusChange = "ClientStatusChange"
    EssenceMessageChange = "EssenceMessageChange"

    MessageReactionsUpdated = "MessageReactionsUpdated"
    ChannelUpdated = "ChannelUpdated"
    ChannelCreated = "ChannelCreated"
    ChannelDestroyed = "ChannelDestroyed"
    GuildChannelRecall = "GuildChannelRecall"

    FriendRequest = "FriendRequest"
    GroupRequest = "GroupRequest"



import base64
import json
import os
import typing as T
from enum import Enum

from pydantic import BaseModel

from ..logger import logger


class ComponentType(Enum):
    Plain = "Plain"
    Face = "Face"
    Record = "Record"
    Video = "Video"
    At = "At"
    RPS = "RPS"  # TODO
    Dice = "Dice"  # TODO
    Shake = "Shake"  # TODO
    Anonymous = "Anonymous"  # TODO
    Share = "Share"
    Contact = "Contact"  # TODO
    Location = "Location"  # TODO
    Music = "Music"
    Image = "Image"
    Reply = "Reply"
    RedBag = "RedBag"
    Poke = "Poke"
    Forward = "Forward"
    Node = "Node"
    Xml = "Xml"
    Json = "Json"
    CardImage = "CardImage"
    TTS = "TTS"
    Unknown = "Unknown"


class BaseMessageComponent(BaseModel):
    type: ComponentType

    def toString(self):
        output = f"[CQ:{self.type.lower()}"
        for k, v in self.__dict__.items():
            if k == "type" or v is None:
                continue
            if k == "_type":
                k = "type"
            if isinstance(v, bool):
                v = 1 if v else 0
            output += ",%s=%s" % (k, str(v).replace("&", "&amp;") \
                                  .replace(",", "&#44;") \
                                  .replace("[", "&#91;") \
                                  .replace("]", "&#93;"))
        output += "]"
        return output

    def toDict(self):
        data = dict()
        for k, v in self.__dict__.items():
            if k == "type" or v is None:
                continue
            if k == "_type":
                k = "type"
            data[k] = v
        return {
            "type": self.type.lower(),
            "data": data
        }


class Plain(BaseMessageComponent):
    type: ComponentType = "Plain"
    text: str
    convert: T.Optional[bool] = True  # 若为 False 则直接发送未转换 CQ 码的消息

    def __init__(self, text: str, convert: bool = True, **_):
        super().__init__(text=text, convert=convert, **_)

    def toString(self):  # 没有 [CQ:plain] 这种东西，所以直接导出纯文本
        if not self.convert:
            return self.text
        return self.text.replace("&", "&amp;") \
            .replace("[", "&#91;") \
            .replace("]", "&#93;")


class Face(BaseMessageComponent):
    type: ComponentType = "Face"
    id: int

    def __init__(self, **_):
        super().__init__(**_)


class Record(BaseMessageComponent):
    type: ComponentType = "Record"
    file: T.Optional[str]
    magic: T.Optional[bool] = False
    url: T.Optional[str]
    cache: T.Optional[bool] = True
    proxy: T.Optional[bool] = True
    timeout: T.Optional[int] = 0
    # 额外
    path: T.Optional[str]

    def __init__(self, file: T.Optional[str], **_):
        for k in _.keys():
            if k == "url":
                logger.warn("Protocol: you should put url into file parameter")
                # Protocol.warn(f"go-cqhttp doesn't support send {self.type} by {k}")
        super().__init__(file=file, **_)

    @staticmethod
    def fromFileSystem(path, **_):
        return Record(file=f"file:///{os.path.abspath(path)}", path=path, **_)

    @staticmethod
    def fromURL(url: str, **_):
        if url.startswith("http://") or url.startswith("https://"):
            return Record(file=url, **_)
        raise Exception("not a valid url")


class Video(BaseMessageComponent):
    type: ComponentType = "Video"
    file: str
    cover: T.Optional[str]
    c: T.Optional[int] = 2
    # 额外
    path: T.Optional[str]

    def __init__(self, file: str, **_):
        for k in _.keys():
            if k == "c" and _[k] not in [2, 3]:
                logger.warn(f"Protocol: {k}={_[k]} doesn't match values")
        super().__init__(file=file, **_)

    @staticmethod
    def fromFileSystem(path, **_):
        return Video(file=f"file:///{os.path.abspath(path)}", path=path, **_)

    @staticmethod
    def fromURL(url: str, **_):
        if url.startswith("http://") or url.startswith("https://"):
            return Video(file=url, **_)
        raise Exception("not a valid url")


class At(BaseMessageComponent):
    type: ComponentType = "At"
    qq: T.Union[int, str]  # 此处str为all时代表所有人
    name: T.Optional[str]

    def __init__(self, **_):
        super().__init__(**_)


class AtAll(At):
    qq: str = "all"

    def __init__(self, **_):
        super().__init__(**_)


class RPS(BaseMessageComponent):  # TODO
    type: ComponentType = "RPS"

    def __init__(self, **_):
        super().__init__(**_)


class Dice(BaseMessageComponent):  # TODO
    type: ComponentType = "Dice"

    def __init__(self, **_):
        super().__init__(**_)


class Shake(BaseMessageComponent):  # TODO
    type: ComponentType = "Shake"

    def __init__(self, **_):
        super().__init__(**_)


class Anonymous(BaseMessageComponent):  # TODO
    type: ComponentType = "Anonymous"
    ignore: T.Optional[bool]

    def __init__(self, **_):
        super().__init__(**_)


class Share(BaseMessageComponent):
    type: ComponentType = "Share"
    url: str
    title: str
    content: T.Optional[str]
    image: T.Optional[str]

    def __init__(self, **_):
        super().__init__(**_)


class Contact(BaseMessageComponent):  # TODO
    type: ComponentType = "Contact"
    _type: str  # type 字段冲突
    id: T.Optional[int]

    def __init__(self, **_):
        for k in _.keys():
            if k == "_type" and _[k] not in ["qq", "group"]:
                logger.warn(f"Protocol: {k}={_[k]} doesn't match values")
        super().__init__(**_)


class Location(BaseMessageComponent):  # TODO
    type: ComponentType = "Location"
    lat: float
    lon: float
    title: T.Optional[str]
    content: T.Optional[str]

    def __init__(self, **_):
        super().__init__(**_)


class Music(BaseMessageComponent):
    type: ComponentType = "Music"
    _type: str
    id: T.Optional[int]
    url: T.Optional[str]
    audio: T.Optional[str]
    title: T.Optional[str]
    content: T.Optional[str]
    image: T.Optional[str]

    def __init__(self, **_):
        for k in _.keys():
            if k == "_type" and _[k] not in ["qq", "163", "xm", "custom"]:
                logger.warn(f"Protocol: {k}={_[k]} doesn't match values")
        super().__init__(**_)


class Image(BaseMessageComponent):
    type: ComponentType = "Image"
    file: T.Optional[str]
    _type: T.Optional[str]
    subType: T.Optional[int]
    url: T.Optional[str]
    cache: T.Optional[bool] = True
    id: T.Optional[int] = 40000
    c: T.Optional[int] = 2
    # 额外
    path: T.Optional[str]

    def __init__(self, file: T.Optional[str], **_):
        for k in _.keys():
            if (k == "_type" and _[k] not in ["flash", "show", None]) or \
                    (k == "c" and _[k] not in [2, 3]):
                logger.warn(f"Protocol: {k}={_[k]} doesn't match values")
        super().__init__(file=file, **_)

    @staticmethod
    def fromURL(url: str, **_):
        if url.startswith("http://") or url.startswith("https://"):
            return Image(file=url, **_)
        raise Exception("not a valid url")

    @staticmethod
    def fromFileSystem(path, **_):
        return Image(file=f"file:///{os.path.abspath(path)}", path=path, **_)

    @staticmethod
    def fromBase64(base64: str, **_):
        return Image(f"base64://{base64}", **_)

    @staticmethod
    def fromBytes(byte: bytes):
        return Image.fromBase64(base64.b64encode(byte).decode())

    @staticmethod
    def fromIO(IO):
        return Image.fromBytes(IO.read())


class Reply(BaseMessageComponent):
    type: ComponentType = "Reply"
    id: int
    text: T.Optional[str]
    qq: T.Optional[int]
    time: T.Optional[int]
    seq: T.Optional[int]

    def __init__(self, **_):
        super().__init__(**_)


class RedBag(BaseMessageComponent):
    type: ComponentType = "RedBag"
    title: str

    def __init__(self, **_):
        super().__init__(**_)


class Poke(BaseMessageComponent):
    type: ComponentType = "Poke"
    qq: int

    def __init__(self, **_):
        super().__init__(**_)


class Forward(BaseMessageComponent):
    type: ComponentType = "Forward"
    id: str

    def __init__(self, **_):
        super().__init__(**_)


class Node(BaseMessageComponent):  # 该 component 仅支持使用 sendGroupForwardMessage 发送
    type: ComponentType = "Node"
    id: T.Optional[int]
    name: T.Optional[str]
    uin: T.Optional[int]
    content: T.Optional[T.Union[str, list]]
    seq: T.Optional[T.Union[str, list]]  # 不清楚是什么
    time: T.Optional[int]

    def __init__(self, content: T.Union[str, list], **_):
        if isinstance(content, list):
            _content = ""
            for chain in content:
                _content += chain.toString()
            content = _content
        super().__init__(content=content, **_)

    def toString(self):
        logger.warn("Protocol: node doesn't support stringify")
        return ""


class Xml(BaseMessageComponent):
    type: ComponentType = "Xml"
    data: str
    resid: T.Optional[int]

    def __init__(self, **_):
        super().__init__(**_)


class Json(BaseMessageComponent):
    type: ComponentType = "Json"
    data: T.Union[str, dict]
    resid: T.Optional[int] = 0

    def __init__(self, data, **_):
        if isinstance(data, dict):
            data = json.dumps(data)
        super().__init__(data=data, **_)


class CardImage(BaseMessageComponent):
    type: ComponentType = "CardImage"
    file: str
    cache: T.Optional[bool] = True
    minwidth: T.Optional[int] = 400
    minheight: T.Optional[int] = 400
    maxwidth: T.Optional[int] = 500
    maxheight: T.Optional[int] = 500
    source: T.Optional[str]
    icon: T.Optional[str]

    def __init__(self, **_):
        super().__init__(**_)

    @staticmethod
    def fromFileSystem(path, **_):
        return CardImage(file=f"file:///{os.path.abspath(path)}", **_)


class TTS(BaseMessageComponent):
    type: ComponentType = "TTS"
    text: str

    def __init__(self, **_):
        super().__init__(**_)


class Unknown(BaseMessageComponent):
    type: ComponentType = "Unknown"
    text: str

    def toString(self):
        return ""


ComponentTypes = {
    "plain": Plain,
    "face": Face,
    "record": Record,
    "video": Video,
    "at": At,
    "rps": RPS,
    "dice": Dice,
    "shake": Shake,
    "anonymous": Anonymous,
    "share": Share,
    "contact": Contact,
    "location": Location,
    "music": Music,
    "image": Image,
    "reply": Reply,
    "redbag": RedBag,
    "poke": Poke,
    "forward": Forward,
    "node": Node,
    "xml": Xml,
    "json": Json,
    "cardimage": CardImage,
    "tts": TTS,
    "unknown": Unknown
}



from collections import namedtuple
from pydantic import BaseModel

InternalEvent = namedtuple("Event", ("name", "body"))

from .enums import ExternalEventTypes

class ExternalEvent(BaseModel):
    type: ExternalEventTypes